package ru.ulstu.method.ftransform;

import org.springframework.stereotype.Component;
import ru.ulstu.datamodel.Model;
import ru.ulstu.datamodel.ts.TimeSeries;
import ru.ulstu.datamodel.ts.TimeSeriesValue;
import ru.ulstu.method.Method;
import ru.ulstu.method.MethodParamValue;
import ru.ulstu.method.MethodParameter;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

@Component
public class FTransform extends Method {
    private final static int NUMBER_OF_FUZZY_VALUES = 3;

    @Override
    protected FTransformModel getModelOfValidTimeSeries(TimeSeries ts,
                                                        List<MethodParamValue> parameters) {
        FTransformModel model = new FTransformModel(ts, parameters);
        List<AComponent> aComponents = generateAComponents(ts, model.getNumberOfCoveredPoints().getIntValue(), model.getAComponents());

        TimeSeries piecewiseLinearTrend = model.getPiecewiseLinearTrend();
        TimeSeries tsModel = model.getTimeSeriesModel();

        for (AComponent aComponent : aComponents) {
            double sum1 = 0;
            double sum2 = 0;
            for (int j = 0; j < ts.getLength(); j++) {
                double membership = aComponent.getValueAtPoint(j);
                sum1 += membership * ts.getNumericValue(j);
                sum2 += membership;
            }
            piecewiseLinearTrend.addValue(ts.getValue(aComponent.getTopInt()), sum1 / sum2);
            tsModel.addValue(ts.getValue(aComponent.getTopInt()), sum1 / sum2);
        }
        return model;
    }

    private List<AComponent> generateAComponents(TimeSeries ts, int numberOfCoveredPoints, List<AComponent> piecewiseLinearTrend) {
        long deltaForTriangle = Math.round(numberOfCoveredPoints / 2.0);
        int currentPoint = 0;
        while (currentPoint < ts.getLength()) {
            int startPoint = (currentPoint == 0)
                    ? 0
                    : piecewiseLinearTrend.get(piecewiseLinearTrend.size() - 1).getTopInt();
            AComponent aComponent = new AComponent(startPoint, currentPoint, (int) (currentPoint + Math.round(numberOfCoveredPoints / 2.0)));
            if (aComponent.getStart() < 0) {
                aComponent.setStart(0);
            }
            if (aComponent.getEnd() > ts.getLength() - 1) {
                aComponent.setEnd(ts.getLength() - 1);
            }

            if (aComponent.getTop() > ts.getLength() - 1) {
                aComponent.setTop(ts.getLength() - 1);
            }

            piecewiseLinearTrend.add(aComponent);
            currentPoint += deltaForTriangle;
        }
        if (piecewiseLinearTrend.get(piecewiseLinearTrend.size() - 1).getEnd() != piecewiseLinearTrend.get(piecewiseLinearTrend.size() - 1).getTop()) {
            AComponent aComponent = new AComponent(piecewiseLinearTrend.get(piecewiseLinearTrend.size() - 1).getTop(),
                    piecewiseLinearTrend.get(piecewiseLinearTrend.size() - 1).getEnd(),
                    piecewiseLinearTrend.get(piecewiseLinearTrend.size() - 1).getEnd());
            piecewiseLinearTrend.add(aComponent);
        }
        return piecewiseLinearTrend;
    }

    @Override
    protected TimeSeries getForecastWithValidParams(Model model, TimeSeries forecast) {
        FTransformModel fTransformModel = (FTransformModel) model;
        double minValue = fTransformModel.getPiecewiseLinearTrend().getValues().stream().map(TimeSeriesValue::getValue).min(Double::compareTo).orElse(0.0);
        double maxValue = fTransformModel.getPiecewiseLinearTrend().getValues().stream().map(TimeSeriesValue::getValue).max(Double::compareTo).orElse(0.0);
        List<AComponent> fuzzyValues = new ArrayList<>();
        double diff = (maxValue - minValue) / NUMBER_OF_FUZZY_VALUES;
        for (int i = 0; i < NUMBER_OF_FUZZY_VALUES; i++) {
            fuzzyValues.add(new AComponent(minValue + i * diff, minValue + i * diff + diff / 2, minValue + i * diff + diff));
        }
        List<AComponent> fuzzyTimeSeries = new ArrayList<>();

        for (int i = 0; i < model.getTimeSeriesModel().getLength(); i++) {
            for (AComponent fuzzyValue : fuzzyValues) {
                if (model.getTimeSeriesModel().getValue(i).getValue() >= fuzzyValue.getStart()
                        && model.getTimeSeriesModel().getValue(i).getValue() <= fuzzyValue.getEnd()) {
                    fuzzyTimeSeries.add(fuzzyValue);
                }
            }
        }


        for (int t = 1; t < forecast.getLength(); t++) {
            forecast.getValues().get(t).setValue(fTransformModel.getPiecewiseLinearTrend().getLastValue().getValue());
        }
        return forecast;
    }

    @Override
    public List<MethodParameter> getAvailableParameters() {
        return FTransformModel.getAvailableParameters();
    }

//    private AComponent findFuzzyValueByPrevious(List<AComponent> fuzzyTimeSeries, AComponent currentFuzzyValue) {
//        Map<AComponent, Integer> mostFrequentValues = new HashMap<>();
//        for (int i = 0; i < fuzzyTimeSeries.size() - 1; i++) {
//            if (fuzzyTimeSeries.get(i).equals(currentFuzzyValue)) {
//                mostFrequentValues.put(fuzzyTimeSeries.get(i + 1), mostFrequentValues.getOrDefault(fuzzyTimeSeries.get(i + 1), 0) + 1);
//            }
//        }
//        mostFrequentValues.entrySet().stream()
//    }

    @Override
    public String getName() {
        return "F - преобразование";
    }
}
